-- ================================================================================================
-- =  ItemPriceTooltip                                                                            =
-- ================================================================================================

-- Import global variables

local assert, error, ipairs, next, pairs, select, tonumber, tostring, type, unpack, pcall, xpcall =
      assert, error, ipairs, next, pairs, select, tonumber, tostring, type, unpack, pcall, xpcall
local getmetatable, setmetatable, rawequal, rawget, rawset, getfenv, setfenv =
      getmetatable, setmetatable, rawequal, rawget, rawset, getfenv, setfenv
local math, string, table = math, string, table
local find, format, gmatch, gsub, tolower, match, toupper, join, split, trim =
      string.find, string.format, string.gmatch, string.gsub, string.lower, string.match, string.upper, string.join, string.split, string.trim
local concat, insert, maxn, remove, sort = table.concat, table.insert, table.maxn, table.remove, table.sort
local max, min, abs, ceil, floor = math.max, math.min, math.abs, math.ceil, math.floor

local AceLibrary = AceLibrary
local MerchantFrame, CreateFrame, GetContainerItemInfo, GetContainerNumSlots, GetContainerItemLink = 
      MerchantFrame, CreateFrame, GetContainerItemInfo, GetContainerNumSlots, GetContainerItemLink

-- No global variables after this!

local _G = getfenv()
do
  local function gerr(t, k) error(format("Reference to global %q", k), 2) end
  setfenv(1, setmetatable({}, {__index = gerr, __newindex = gerr}))
end


-- ================================================================================================
-- = Localization                                                                                 =
-- ================================================================================================

local L = AceLibrary("AceLocale-2.2"):new("ItemPriceTooltip")

L:RegisterTranslations("enUS", function() return {
    ["Sells for"] = true,
  }
end)

L:RegisterTranslations("deDE", function() return {
    ["Sells for"] = "Verkauf f\195\188r",
  }
end)

L:RegisterTranslations("frFR", function() return {
    ["Sells for"] = "Vendu pour",
  }
end)

L:RegisterTranslations("koKR", function() return {
    ["Sells for"] = "판매 가격",
  }
end)

L:RegisterTranslations("zhTW", function() return {
    ["Sells for"] = "賣出價",
  }
end)


-- ================================================================================================
-- =  Initialization (addon, libraries, etc.)                                                     =
-- ================================================================================================

local mixins = {"AceHook-2.1", "AceEvent-2.0", "AceDB-2.0", "AceConsole-2.0"}
local hasAceDebug = AceLibrary:HasInstance("AceDebug-2.0")
if hasAceDebug then insert(mixins, "AceDebug-2.0") end
  
local ItemPriceTooltip = AceLibrary("AceAddon-2.0"):new(unpack(mixins))

ItemPriceTooltip:RegisterDB("ItemPriceTooltipDB")

local function noop() end
local dbg, debug, ldebug
if hasAceDebug then
  function dbg() return ItemPriceTooltip:IsDebugging() end
  function debug(...) return ItemPriceTooltip:Debug(...) end
  function ldebug(level, ...) return ItemPriceTooltip:LevelDebug(level, ...) end
  -- ItemPriceTooltip:SetDebugging(true)
  -- ItemPriceTooltip:SetDebugLevel(1)
else
  dbg, debug, ldebug = noop, noop, noop
end

-- Libraries
local ItemPrice = AceLibrary("ItemPrice-1.0")
local Abacus = AceLibrary("Abacus-2.0")


local function linkToId(link)
  return assert(tonumber(match(link, "item:(%d+):")))
end

-- ================================================================================================
-- =  Tooltip hooking functions                                                                   =
-- ================================================================================================

local Hooks = {}
do
  local SellsForText1 = format("|cff7fff7f%s|r", L["Sells for"])
  local SellsForTextN = format("%s |cffffffaa(|r|cffffffff%%d|r |cffaaaaaax|r %%s|cffffffaa)|r", SellsForText1)

  local function formatPriceText(price, count)
    local pstr1 = Abacus:FormatMoneyShort(price, true)
    local pstrN = Abacus:FormatMoneyFull(price * count, true)
    if count == 1 then return SellsForText1, pstrN end
    return format(SellsForTextN, count, pstr1), pstrN
  end

  local function addPrice(tooltip, count)
    if dbg() then ldebug(3, "addPrice(%s, %s)", tostring(tooltip), tostring(count)) end
    count = count or 1
    local _, item = tooltip:GetItem()
    if not item then return end
    local itemId = linkToId(item)
    local price = ItemPriceTooltip:getPriceDB()[itemId] or ItemPrice:GetPrice(itemId)
    if not price then return end
    local left, right = formatPriceText(price, count)
    tooltip:AddDoubleLine(left, right)
    if tooltip:IsShown() then tooltip:Show() end
  end

  function Hooks:SetAction(id)
    if not self:GetItem() then return end
    local count = (_G.IsConsumableAction(id) or _G.IsStackableAction(id)) and _G.GetActionCount(id)
    addPrice(self, count)
  end

  function Hooks:SetAuctionItem(type, index)
    local _, _, count = _G.GetAuctionItemInfo(type, index)
    addPrice(self, count)
  end

  function Hooks:SetAuctionSellItem()
    local _, _, count = _G.GetAuctionSellItemInfo()
    addPrice(self, count)
  end

  function Hooks:SetBagItem(bag, slot)
    if MerchantFrame:IsShown() then return end
    local _, count = _G.GetContainerItemInfo(bag, slot)
    addPrice(self, count)
  end

  function Hooks:SetCraftItem(skill, slot)
    local count = 1
    if slot then
      count = select(3, _G.GetCraftReagentInfo(skill, slot))
    end
    addPrice(self, count)
  end

  function Hooks:SetHyperlink(link)
    addPrice(self, 1)
  end

  function Hooks:SetInboxItem(index)
    local _, _, count = _G.GetInboxItem(index)
    addPrice(self, count)
  end

  local showBagTooltips = false -- Might make this an option later.

  function Hooks:SetInventoryItem(unit, slot)
    if type(slot) ~= "number" or slot < 0 then return end
    if not showBagTooltips and (slot >= 20 and slot <= 23 or slot >= 68) then return end
    local count = 1
    if slot < 20 or slot > 39 and slot < 68 then
      count = _G.GetInventoryItemCount(unit, slot)
    end
    addPrice(self, count)
  end

  function Hooks:SetLootItem(slot)
    local _, _, count = _G.GetLootSlotInfo(slot)
    addPrice(self, count)
  end

  function Hooks:SetLootRollItem(rollID)
    local _, _, count = _G.GetLootRollItemInfo(rollID)
    addPrice(self, count)
  end

  function Hooks:SetMerchantCostItem(index, item) -- FIXME: Needs testing!
    local _, count = _G.GetMerchantItemCostItem(index, item)
    addPrice(self, count)
  end

  function Hooks:SetMerchantItem(slot)
    local _, _, _, count = _G.GetMerchantItemInfo(slot)
    addPrice(self, count)
  end

  function Hooks:SetQuestItem(type, slot)
    local _, _, count = _G.GetQuestItemInfo(type, slot)
    addPrice(self, count)
  end

  function Hooks:SetQuestLogItem(type, index)
    local _, _, count = _G.GetQuestLogRewardInfo(index)
    addPrice(self, count)
  end

  function Hooks:SetSendMailItem()
    local _, _, count = _G.GetSendMailItem()
    addPrice(self, count)
  end

  function Hooks:SetSocketedItem() -- FIXME: Needs testing!
    addPrice(self, 1)
  end

  function Hooks:SetExistingSocketGem() -- FIXME: Needs testing!
    addPrice(self, 1)
  end

  function Hooks:SetSocketGem() -- FIXME: Needs testing!
    addPrice(self, 1)
  end

  function Hooks:SetTradePlayerItem(index)
    local _, _, count = _G.GetTradePlayerItemInfo(index)
    addPrice(self, count)
  end

  function Hooks:SetTradeSkillItem(skill, slot)
    local count = 1
    if slot then
      count = select(3, _G.GetTradeSkillReagentInfo(skill, slot))
    end
    addPrice(self, count)
  end

  function Hooks:SetTradeTargetItem(index)
    local _, _, count = _G.GetTradeTargetItemInfo(index)
    addPrice(self, count)
  end

  -- Hook into Link's tooltips (via the event "Links_LinkOpened")
  -- See: http://www.wowace.com/wiki/Links
  function ItemPriceTooltip:Links_LinkOpened(_, frame)
    addPrice(frame, 1)
  end
end


-- ================================================================================================
-- = Vendor scanning                                                                              =
-- ================================================================================================

local vendor_scan
do
  local tooltip_money
  local tooltip = CreateFrame("GameTooltip")
  tooltip:SetScript("OnTooltipAddMoney", function(_, money) tooltip_money = money end)
  
  local function get_price_at_vendor(bag, slot)
    local _, count = GetContainerItemInfo(bag, slot)
    if not count or count < 1 then return end
    tooltip_money = nil
    local _, repairCost = tooltip:SetBagItem(bag, slot)
    if type(repairCost) == "number" and repairCost > 0 then return end -- Chickening out - don't want repair cost to affect saved prices!
    if tooltip_money then
      if count > 1 then return tooltip_money / count end
      return tooltip_money
    end
  end

  local function scan_slot(new_prices, bag, slot)
	  local link = GetContainerItemLink(bag, slot)
	  if not link then return end
    local itemId = linkToId(link)
    local new_price = new_prices[itemId]
    if new_price then return end
    new_price = get_price_at_vendor(bag, slot)
    if not new_price then return end
    local lib_price = ItemPrice:GetPrice(itemId)
    if new_price == lib_price then return end
    new_prices[itemId] = new_price
    if dbg() then debug("Price: %s = %s (%s)", link, tostring(new_price), tostring(lib_price)) end
  end

  function vendor_scan(self)
    if not MerchantFrame:IsShown() then return end
    if _G.InRepairMode() then return end
    debug("Scanning vendor")
    local new_prices = self:getPriceDB()
  	for bag = 0, _G.NUM_BAG_FRAMES do
  		for slot = 1, GetContainerNumSlots(bag) do
  		  scan_slot(new_prices, bag, slot)
      end
    end
  end
end


-- ================================================================================================
-- =  Addon methods                                                                               =
-- ================================================================================================

function ItemPriceTooltip:OnInitialize()
  self:RegisterChatCommand({ "/itempricetooltip", "/ipt" })
  local db_name = ItemPrice:GetLibraryVersion()
  self.ip_db = self:AcquireDBNamespace(db_name).account
  local defaults = {}
  defaults.prices = {}
  self:RegisterDefaults(db_name, 'account', defaults)
end

function ItemPriceTooltip:getPriceDB()
  local ip_db = self.ip_db
  local db_version = ip_db.version
  local _, lib_version = ItemPrice:GetLibraryVersion()
  if db_version == lib_version then return ip_db.prices end
  local prices = {}
  for id, price in pairs(ip_db.prices) do
    if ItemPrice:GetPrice(id) ~= price then
      prices[id] = price
    end
  end
  ip_db.version = lib_version
  ip_db.prices = prices
  return prices
end

function ItemPriceTooltip:OnEnable()
  self:installHooks(_G.GameTooltip, Hooks)
  self:installHooks(_G.ItemRefTooltip, Hooks)

  self:RegisterEvent("MERCHANT_SHOW", 0.5)

  self:RegisterEvent("Links_LinkOpened") -- See: http://www.wowace.com/wiki/Links
  
  self:getPriceDB()
end

function ItemPriceTooltip:installHooks(tooltip, hooks)
  for name, func in pairs(hooks) do
    if type(tooltip[name]) == "function" then
      self:SecureHook(tooltip, name, func)
    end
  end
end

function ItemPriceTooltip:MERCHANT_SHOW()
  vendor_scan(self)
end

